@@ -1,4 +1,4 @@
-#! /usr/bin/perl -w
+#!/usr/bin/perl -w
 # -*- tab-width: 8; indent-tabs-mode: t; cperl-indent-level: 4 -*-
 
 # uscan: This program looks for watchfiles and checks upstream ftp sites
@@ -46,6 +46,8 @@ BEGIN {
 	}
     }
 }
+use Dpkg::Control::Hash;
+
 my $CURRENT_WATCHFILE_VERSION = 3;
 
 my $progname = basename($0);
@@ -53,7 +55,7 @@ my $modified_conf_msg;
 my $opwd = cwd();
 
 my $haveSSL = 1;
-eval { require Crypt::SSLeay; };
+eval { require LWP::Protocol::https; };
 if ($@) {
     $haveSSL = 0;
 }
@@ -72,6 +74,7 @@ sub uscan_die (@);
 sub dehs_output ();
 sub quoted_regex_replace ($);
 sub safe_replace ($$);
+sub get_main_source_dir($$$$);
 
 sub usage {
     print <<"EOF";
@@ -138,6 +141,8 @@ Options:
     --no-conf, --noconf
                    Don\'t read devscripts config files;
                    must be the first option given
+    --no-exclusion no automatic exclusion of files mentioned in
+                   debian/copyright field Files-Excluded
     --help         Show this message
     --version      Show version information
 
@@ -180,6 +185,7 @@ my $dehs_start_output = 0;
 my $pkg_report_header = '';
 my $timeout = 20;
 my $user_agent_string = 'Debian uscan ###VERSION###';
+my $exclusion = 1;
 
 if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     $modified_conf_msg = "  (no configuration files read)";
@@ -196,6 +202,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 		       'USCAN_DEHS_OUTPUT' => 'no',
 		       'USCAN_USER_AGENT' => '',
 		       'USCAN_REPACK' => 'no',
+		       'USCAN_EXCLUSION' => 'yes',
 		       'DEVSCRIPTS_CHECK_DIRNAME_LEVEL' => 1,
 		       'DEVSCRIPTS_CHECK_DIRNAME_REGEX' => 'PACKAGE(-.+)?',
 		       );
@@ -233,6 +240,8 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'USCAN_DEHS_OUTPUT'}='no';
     $config_vars{'USCAN_REPACK'} =~ /^(yes|no)$/
 	or $config_vars{'USCAN_REPACK'}='no';
+    $config_vars{'USCAN_EXCLUSION'} =~ /^(yes|no)$/
+	or $config_vars{'USCAN_EXCLUSION'}='yes';
     $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'} =~ /^[012]$/
 	or $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'}=1;
 
@@ -263,7 +272,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 # Now read the command line arguments
 my $debug = 0;
 my ($opt_h, $opt_v, $opt_destdir, $opt_download, $opt_force_download,
-    $opt_report, $opt_passive, $opt_symlink, $opt_repack);
+    $opt_report, $opt_passive, $opt_symlink, $opt_repack, $opt_exclusion);
 my ($opt_verbose, $opt_level, $opt_regex, $opt_noconf);
 my ($opt_package, $opt_uversion, $opt_watchfile, $opt_dehs, $opt_timeout);
 my $opt_download_version;
@@ -295,6 +304,7 @@ GetOptions("help" => \$opt_h,
 	   "useragent=s" => \$opt_user_agent,
 	   "noconf" => \$opt_noconf,
 	   "no-conf" => \$opt_noconf,
+	   "exclusion!" => \$opt_exclusion,
 	   "download-current-version" => \$opt_download_current_version,
 	   )
     or die "Usage: $progname [options] [directories]\nRun $progname --help for more details\n";
@@ -318,6 +328,7 @@ $timeout = 20 unless defined $timeout and $timeout > 0;
 $symlink = $opt_symlink if defined $opt_symlink;
 $verbose = $opt_verbose if defined $opt_verbose;
 $dehs = $opt_dehs if defined $opt_dehs;
+$exclusion = $opt_exclusion if defined $opt_exclusion;
 $user_agent_string = $opt_user_agent if defined $opt_user_agent;
 $download_version = $opt_download_version if defined $opt_download_version;
 
@@ -875,7 +886,7 @@ sub process_watchline ($$$$$$)
     # Devscripts::Versort::upstream_versort
     if ($site =~ m%^http(s)?://%) {
 	if (defined($1) and !$haveSSL) {
-	    uscan_die "$progname: you must have the libcrypt-ssleay-perl package installed\nto use https URLs\n";
+	    uscan_die "$progname: you must have the liblwp-protocol-https-perl package installed\nto use https URLs\n";
 	}
 	print STDERR "$progname debug: requesting URL $base\n" if $debug;
 	$request = HTTP::Request->new('GET', $base, $headers);
@@ -1341,7 +1352,7 @@ EOF
     # Download newer package
     if ($upstream_url =~ m%^http(s)?://%) {
 	if (defined($1) and !$haveSSL) {
-	    uscan_die "$progname: you must have the libcrypt-ssleay-perl package installed\nto use https URLs\n";
+	    uscan_die "$progname: you must have the liblwp-protocol-https-perl package installed\nto use https URLs\n";
 	}
 	# substitute HTML entities
 	# Is anything else than "&amp;" required?  I doubt it.
@@ -1447,7 +1458,7 @@ EOF
 	$newfile_base = $newfile_base_gz;
     }
 
-    if ($repack and $newfile_base =~ /^(.*)\.zip$/) {
+    if ($repack and $newfile_base =~ /^(.*)\.(zip|jar)$/) {
 	print "-- Repacking from zip to .tar.gz\n" if $verbose;
 
 	system('command -v unzip >/dev/null 2>&1') >> 8 == 0
@@ -1459,12 +1470,12 @@ EOF
 	my $hidden = ".[!.]*";
 	my $absdestdir = abs_path($destdir);
 	system('unzip', '-q', '-a', '-d', $tempdir, "$destdir/$newfile_base") == 0
-	  or uscan_die("Repacking from zip to tar.gz failed (could not unzip)\n");
+	  or uscan_die("Repacking from zip or jar to tar.gz failed (could not unzip)\n");
 	if (defined glob("$tempdir/$hidden")) {
 	    $globpattern .= " $hidden";
 	}
 	system("cd $tempdir; GZIP='-n -9' tar --owner=root --group=root --mode=a+rX -czf \"$absdestdir/$newfile_base_gz\" $globpattern") == 0
-	  or uscan_die("Repacking from zip to tar.gz failed (could not create tarball)\n");
+	  or uscan_die("Repacking from zip or jar to tar.gz failed (could not create tarball)\n");
 	unlink "$destdir/$newfile_base";
 	$newfile_base = $newfile_base_gz;
     }
@@ -1480,6 +1491,54 @@ EOF
 	}
     }
 
+    my $excludesuffix = '+dfsg';
+    if ($exclusion) {
+	my $data = Dpkg::Control::Hash->new();
+	$data->load('debian/copyright');
+	my $okformat = qr'http://www.debian.org/doc/packaging-manuals/copyright-format/[.\d]+';
+	if ($data->{'format'} =~ m{^$okformat/?$} and $data->{'files-excluded'} ) {
+	    my $tempdir = tempdir ( "uscanXXXX", TMPDIR => 1, CLEANUP => 1 );
+	    my $globpattern = "*";
+	    my $hidden = ".[!.]*";
+	    if (defined glob("$tempdir/$hidden")) {
+		$globpattern .= " $hidden";
+	    }
+	    my $absdestdir = abs_path($destdir);
+	    unless (system("cd $tempdir; tar -xaf \"$absdestdir/$newfile_base\" 2>/dev/null") == 0) {
+		print "-- $newfile_base is no tarball.  Try unzip.\n" if $verbose;
+		# try unzip if tar fails - we do want to do something sensible even if no --repack was specified
+		system('command -v unzip >/dev/null 2>&1') >> 8 == 0
+		    or uscan_die("unzip binary not found. This would serve as fallback because tar just failed.\n");
+		system('unzip', '-q', '-a', '-d', $tempdir, "$destdir/$newfile_base") == 0
+		    or uscan_die("Repacking from zip or jar to tar.gz failed (could not unzip)\n");
+	    }
+	    my $main_source_dir = get_main_source_dir($tempdir, $pkg, $newversion, $excludesuffix);
+	    unless ( -d $main_source_dir ) {
+		print STDERR "Error: $main_source_dir is no directory";
+	    }
+	    my $nfiles_before = `find $main_source_dir | wc -l`;
+	    foreach (grep {/\//} split /\s+/, $data->{"files-excluded"}) {
+		# delete trailing '/' because otherwise find -path will fail
+		s?/+$?? ;
+		# use rm -rf to enable deleting non-empty directories
+		`find $main_source_dir -path "$main_source_dir/$_" | xargs rm -rf`;
+	    };
+	    foreach (grep {/^[^\/]+$/} split /\s+/, $data->{"files-excluded"}) {
+		`find $main_source_dir -type f -name $_ -delete`;
+	    };
+	    my $nfiles_after = `find $main_source_dir | wc -l`;
+	    if ( $nfiles_before == $nfiles_after ) {
+		print "-- Source tree remains identical - no need for repacking.\n" if $verbose;
+	    } else {
+		my $suffix = 'gz' ;
+		my $newfile_base_dfsg = "${pkg}_${newversion}${excludesuffix}.orig.tar.$suffix" ;
+		system("cd $tempdir; GZIP='-n -9' tar --owner=root --group=root --mode=a+rX -czf \"$absdestdir/$newfile_base_dfsg\" $globpattern") == 0
+		    or die("Excluding files failed (could not create tarball)\n");
+		$symlink = 'files-excluded' # prevent symlinking or renaming
+	    }
+	}
+    }
+
     my @renames = (
 	[qr/\.(tar\.gz|tgz)$/, 'gz'],
 	[qr/\.(tar\.bz2|tbz2?)$/, 'bz2'],
@@ -1506,6 +1565,8 @@ EOF
 		print "    and symlinked $renamed_base to it\n";
 	    } elsif ($symlink eq 'rename') {
 		print "    and renamed it as $renamed_base\n";
+	    } elsif ($symlink eq 'files-excluded') {
+		print "    and removed files from it in ${pkg}_${newversion}${excludesuffix}.orig.tar.$suffix\n";
 	    }
 	} elsif ($dehs) {
 	    my $msg = "Successfully downloaded updated package $newfile_base";
@@ -1514,6 +1575,8 @@ EOF
 		$msg .= " and symlinked $renamed_base to it";
 	    } elsif ($symlink eq 'rename') {
 		$msg .= " and renamed it as $renamed_base";
+	    } elsif ($symlink eq 'files-excluded') {
+		$msg .= " and removed files from it in ${pkg}_${newversion}${excludesuffix}.orig.tar.$suffix\n";
 	    } else {
 		$dehs_tags{'target'} = $newfile_base;
 	    }
@@ -1524,6 +1587,8 @@ EOF
 		print "    and symlinked $renamed_base to it\n";
 	    } elsif ($symlink eq 'rename') {
 		print "    and renamed it as $renamed_base\n";
+	    } elsif ($symlink eq 'files-excluded') {
+		print "    and removed files from it in ${pkg}_${newversion}${excludesuffix}.orig.tar.$suffix\n";
 	    }
 	}
 	last;
@@ -1603,7 +1668,7 @@ sub newest_dir ($$$$$) {
 
     if ($site =~ m%^http(s)?://%) {
 	if (defined($1) and !$haveSSL) {
-	    uscan_die "$progname: you must have the libcrypt-ssleay-perl package installed\nto use https URLs\n";
+	    uscan_die "$progname: you must have the liblwp-protocol-https-perl package installed\nto use https URLs\n";
 	}
 	print STDERR "$progname debug: requesting URL $base\n" if $debug;
 	$request = HTTP::Request->new('GET', $base);
@@ -2066,3 +2131,52 @@ sub safe_replace($$) {
 	return 1;
     }
 }
+
+sub get_main_source_dir($$$$) {
+    my ($tempdir, $pkg, $newversion, $excludesuffix) = @_;
+    my $fcount = 0;
+    my $main_source_dir = '';
+    my $any_dir = '';
+    opendir DIR, $tempdir or uscan_die "opendir $tempdir: $!";
+    my @files = readdir DIR;
+    closedir DIR;
+    foreach my $file (@files) {
+	unless ($file =~ /^\.\.?$/) {
+	    $fcount++;
+	    if (-d $tempdir.'/'.$file) {
+		$any_dir = $tempdir . '/' . $file;
+		# check whether there is some dir in upstream source which looks reasonable
+		# If such dir exists, we do not try to undirty the directory structure
+		$main_source_dir = $any_dir if $file =~ /^$pkg\w*$newversion$/i;
+	    }
+	}
+    }
+    if ($fcount == 1 and $main_source_dir) {
+	return $main_source_dir;
+    }
+    if ($fcount == 1 and $any_dir) {
+	# Unusual base dir in tarball - should be rather something like ${pkg}-${newversion}
+	$main_source_dir = $tempdir . '/' . $pkg . '-' . $newversion . $excludesuffix . '.orig';
+	move($any_dir, $main_source_dir) or uscan_die("Unable to move $any_dir directory $main_source_dir\n");
+	return $main_source_dir;
+    }
+    print "-- Dirty tarball found.\n" if $verbose;
+    if ($main_source_dir) { # if tarball is dirty but does contain a $pkg-$newversion dir we will not undirty but leave it as is
+	print "-- No idea how to create proper tarball structure - leaving as is.\n" if $verbose;
+	return $tempdir;
+    }
+    print "-- Move files to subdirectory $pkg-$newversion.\n" if $verbose;
+    $main_source_dir = $tempdir . '/' . $pkg . '-' . $newversion . $excludesuffix . '.orig';
+    mkdir($main_source_dir) or uscan_die("Unable to create temporary source directory $main_source_dir\n");
+    foreach my $file (@files) {
+	unless ($file =~ /^\.\.?/) {
+	    if ( -d "${tempdir}/$file" ) {
+                # HELP: why can't perl move not move directories????
+                system( "mv ${tempdir}/$file $main_source_dir" ) ;
+            } else {
+                move("${tempdir}/$file", $main_source_dir) or die("Unable to move ${tempdir}/$file directory $main_source_dir\n");
+            }
+        }
+    }
+    return $main_source_dir;
+}
